"""
Python code for reading CHIME/FRB msgpack data.
"""

import msgpack
import numpy as np


class AssembledChunk:
    def __init__(self, msgpacked_chunk, debug=False):
        c = msgpacked_chunk
        # print('header', c[0])
        version = c[1]
        assert version in [1, 2]
        if version == 1:
            assert len(c) == 17
        if version == 2:
            assert len(c) == 21
        self.version = version
        compressed = c[2]
        compressed_size = c[3]

        if debug:
            print("data version", version)
            print("compressed? ", compressed)
            print("compressed size", compressed_size)

        self.beam = c[4]
        self.nupfreq = c[5]
        self.nt_per_packet = c[6]
        self.fpga_counts_per_sample = c[7]
        self.nt_coarse = c[8]
        self.nscales = c[9]
        self.ndata = c[10]
        self.fpga0 = c[11]
        self.fpgaN = c[12]
        self.binning = c[13]
        self.nt = self.nt_coarse * self.nt_per_packet

        scales = c[14]
        offsets = c[15]
        data = c[16]

        # version 2: extra arguments
        self.frame0_nano = 0
        self.nrfifreq = 0
        self.has_rfi_mask = False
        self.rfi_mask = None
        if self.version == 2:
            self.frame0_nano = c[17]
            self.nrfifreq = c[18]
            self.has_rfi_mask = c[19]
            mask = c[20]
            # to numpy
            mask = np.fromstring(mask, dtype=np.uint8)
            mask = mask.reshape((self.nrfifreq, self.nt // 8))
            # Expand mask!
            self.rfi_mask = np.zeros((self.nrfifreq, self.nt), bool)
            for i in range(8):
                self.rfi_mask[:, i::8] = (mask & (1 << i)) > 0

        # Convert to numpy arrays
        self.scales = np.fromstring(scales, dtype="<f4")
        self.offsets = np.fromstring(offsets, dtype="<f4")
        self.scales = self.scales.reshape((-1, self.nt_coarse))
        self.offsets = self.offsets.reshape((-1, self.nt_coarse))
        self.data = np.frombuffer(data, dtype=np.uint8)
        self.data = self.data.reshape((-1, self.nt))

    def __str__(self):
        if self.has_rfi_mask:
            h, w = self.rfi_mask.shape
            masked = np.sum(self.rfi_mask == 0)
            rfistr = "yes, %i freqs, %i%% masked" % (
                self.nrfifreq,
                int(100.0 * masked / (h * w)),
            )
        else:
            rfistr = "no"
        return "AssembledChunk: beam %i, nt %i, fpga0 %i, rfi %s" % (
            self.beam,
            self.nt,
            self.fpga0,
            rfistr,
        )

    def decode(self, debug=False):
        # Returns (intensities,weights) as floating-point
        nf = self.data.shape[1]

        if debug:
            print("Data shape:", self.data.shape)
            print("Scales shape:", self.scales.shape)
            print("nupfreq:", self.nupfreq)
            print("nt_per_packet:", self.nt_per_packet)
            print("nf: ", nf)

        intensities = (
            self.offsets.repeat(self.nupfreq, axis=0).repeat(self.nt_per_packet, axis=1)
            + self.data
            * self.scales.repeat(self.nupfreq, axis=0).repeat(
                self.nt_per_packet, axis=1
            )
        ).astype(np.float32)

        weights = ((self.data > 0) * (self.data < 255)) * np.float32(1.0)

        return intensities, weights

    def time_start(self):
        # Nanoseconds per FPGA count
        fpga_nano = 2560
        return 1e-9 * (
            self.frame0_nano + self.fpga_counts_per_sample * fpga_nano * self.fpga0
        )

    def time_end(self):
        # Nanoseconds per FPGA count
        fpga_nano = 2560
        return 1e-9 * (
            self.frame0_nano
            + self.fpga_counts_per_sample * fpga_nano * (self.fpga0 + self.fpgaN)
        )


def read_msgpack_file(fn):
    f = open(fn, "rb")
    m = msgpack.unpackb(f.read())
    return AssembledChunk(m)



from csv import DictReader
from json import dumps
from logging import getLogger

log = getLogger(__name__)


def fits_to_dataframe(filename):
    try:
        import pandas as pd
    except ImportError as error:
        log.error(error)
        log.error("Run 'pip install cfod[pandas]'")
        raise error
    columns = [
        "tns_name",
        "previous_name",
        "repeater_name",
        "ra",
        "ra_err",
        "ra_notes",
        "dec",
        "dec_err",
        "dec_notes",
        "gl",
        "gb",
        "exp_up",
        "exp_up_err",
        "exp_up_notes",
        "exp_low",
        "exp_low_err",
        "exp_low_notes",
        "bonsai_snr",
        "bonsai_dm",
        "low_ft_68",
        "up_ft_68",
        "low_ft_95",
        "up_ft_95",
        "snr_fitb",
        "dm_fitb",
        "dm_fitb_err",
        "dm_exc_ne2001",
        "dm_exc_ymw16",
        "bc_width",
        "scat_time",
        "scat_time_err",
        "flux",
        "flux_err",
        "flux_notes",
        "fluence",
        "fluence_err",
        "fluence_notes",
        "sub_num",
        "mjd_400",
        "mjd_400_err",
        "mjd_inf",
        "mjd_inf_err",
        "width_fitb",
        "width_fitb_err",
        "sp_idx",
        "sp_idx_err",
        "sp_run",
        "sp_run_err",
        "high_freq",
        "low_freq",
        "peak_freq",
        "excluded_flag",
    ]
    return pd.read_csv(filename, usecols=columns)


def csv_to_list(filename: str) -> list:
    reader = DictReader(open(filename))
    data = []
    for datapoint in reader:
        data.append(datapoint)
    return data


def csv_to_dict(filename: str) -> dict:
    reader = DictReader(open(filename))
    data = {}
    for datapoint in reader:
        data[datapoint["tns_name"]] = datapoint
    return data


def csv_to_json(filename: str) -> str:
    dictionary = csv_to_dict(filename)
    return dumps(dictionary)



#!/bin/python

"""Script to convert a list of CHIME/FRB msgpack files to a filterbank file.

Ziggy Pleunis, ziggy.pleunis@physics.mcgill.ca

"""

import glob
import os

import click
import numpy as np

from cfod.analysis.filterbank import filterbank
from cfod.analysis.intensity import chime_intensity

# 16K channels, 1024 samples, intensity + weights, 32-bits = 4 bytes
MSGPACK_SIZE = 16e3 * 1024 * 2 * 4
SCRUNCH_FACTORS = [
    1,
    2,
    4,
    8,
    16,
    32,
    64,
    128,
    256,
    512,
    1024,
    2048,
    4096,
    8192,
    16384,
]


def filterbank_header(obsglob, dt, ftop, fbottom, nchan, source):
    """Create a filterbank header.

    Returns
    -------
    fil_header : dict
        Header in the filterbank format.

    Notes
    -----
    Basic implementation with a lot of unfilled header variables.

    """
    fil_header = {}
    fil_header["telescope_id"] = 20  # number for CHIME
    fil_header["machine_id"] = 20  # number for CHIME/FRB backend
    fil_header["data_type"] = 1  # filterbank
    # PRESTO and sigproc read at most 80 characters for header messages
    fil_header["rawdatafile"] = obsglob[:80]
    fil_header["source_name"] = source
    fil_header["barycentric"] = 0
    fil_header["pulsarcentric"] = 0
    fil_header["az_start"] = 0.0  # degrees
    fil_header["za_start"] = 0.0  # degrees
    fil_header["src_raj"] = 0.0  # [hhmmss.s]
    fil_header["src_dej"] = 0.0  # [ddmmss.s]
    fil_header["tstart"] = 0.0  # MJD
    fil_header["tsamp"] = dt  # s
    fil_header["nbits"] = 32
    fil_header["nbeams"] = 1
    fil_header["ibeam"] = 0
    # first channel `fch1` in sigproc is the highest frequency
    # `foff` is negative to signify this
    channel_bandwidth = float(np.abs(ftop - fbottom)) / nchan
    fil_header["fch1"] = ftop - channel_bandwidth / 2.0  # MHz
    fil_header["foff"] = -1.0 * channel_bandwidth  # MHz
    fil_header["nchans"] = nchan
    fil_header["nifs"] = 1

    return fil_header


def average(a, axis=None, weights=None):
    """Same as np.average, EXCEPT, if weights are all zero, it will
    return zero in the average without giving an error.

    Parameters
    ----------
    a : array_like
        The array to take the average off.
    axis : int, optional
        The axis to calculate the average over.
    weights : array_like, optional
        Data weights. Use equal weighting by default.

    Returns
    -------
    float
        The array average over the provided axis.

    """
    assert a.ndim >= 2

    if weights is None:
        weights = np.ones(a.shape, dtype=np.float)

    weight_sums = np.sum(weights, axis=axis)

    non_zero = np.where(weight_sums > 0)
    avg = np.zeros_like(weight_sums)
    avg[non_zero] = np.sum(a * weights, axis=axis)[non_zero] / weight_sums[non_zero]

    return avg, weight_sums


def delay_from_dm(dm, freq_emitted):
    """Return the delay in seconds caused by dispersion in the
    interstellar medium.

    Parameters
    ----------
    dm : float
        Dispersion measure, in pc cm-3.
    freq_emitted : float
        Observing frequency, in MHz.

    Returns
    -------
    float
        Dispersive delay, in seconds.

    """
    if isinstance(freq_emitted, float):
        if freq_emitted > 0.0:
            return dm / (0.000241 * freq_emitted * freq_emitted)
        else:
            return 0.0
    else:
        return np.where(
            freq_emitted > 0.0,
            dm / (0.000241 * freq_emitted * freq_emitted),
            0.0,
        )


def convert_chunk(msg_chunk, fscrunch=1, subdm=None):
    # load a list of msgpack files
    (
        intensity,
        weights,
        fpga0s,
        fpgaNs,
        binning,
        rfi_masks,
        frame0_nanos,
    ) = chime_intensity.unpack_datafiles(msg_chunk)

    dt = chime_intensity.dt * binning

    if fscrunch not in SCRUNCH_FACTORS:
        fscrunch = 4
        print("WARNING fscrunch is not a factor of 2.. setting it to 4!")

    nsub = 16384 / fscrunch
    # give warning when nchan > 4096
    if nsub > 4096:
        print(
            "WARNING sigproc spectrum lengths are capped at 4096 "
            + "channels by default, need to update `reader.h` and "
            + "`header.h` and recompile before reading this file!"
        )

    nchan = intensity.shape[0]

    nchan_per_sub = nchan / nsub

    # update frequency channel width
    df = nchan_per_sub * chime_intensity.df

    if nchan != nsub:
        old_frequencies = np.arange(
            chime_intensity.fbottom, chime_intensity.ftop, chime_intensity.df
        )
        old_center_frequencies = old_frequencies + chime_intensity.df / 2.0

        # calculate subband frequencies for subband dedispersion
        new_frequencies = np.arange(chime_intensity.fbottom, chime_intensity.ftop, df)
        new_center_frequencies = new_frequencies + df / 2.0

        # dedisperse channels *within* subbands to `subdm`
        if subdm is not None:
            # compute delays
            rel_delays = delay_from_dm(subdm, new_center_frequencies)
            delays = delay_from_dm(subdm, old_center_frequencies)
            # relative delays
            rel_delays = delays - rel_delays.repeat(nchan_per_sub)
            rel_bindelays = np.round(rel_delays / dt).astype("int")
            # shift channels
            for ii in range(nchan):
                # rotate channels
                intensity[ii, :] = np.roll(intensity[ii, :], -rel_bindelays[ii], axis=0)
                weights[ii, :] = np.roll(weights[ii, :], -rel_bindelays[ii], axis=0)

                # zero out rotated values in the weights array
                if rel_bindelays[ii] > 0:
                    weights[ii, -rel_bindelays[ii] :] = 0.0
                elif rel_bindelays[ii] < 0:
                    weights[ii, : -rel_bindelays[ii]] = 0.0

        # subband
        intensity = np.array(
            [
                average(sub, axis=0, weights=sub_weights)[0]
                for sub, sub_weights in zip(
                    np.vsplit(intensity, nsub), np.vsplit(weights, nsub)
                )
            ]
        )
        weights = np.array(
            [np.mean(weights, axis=0) for weights in np.vsplit(weights, nsub)]
        )

    return intensity, weights, dt, df


def msgpack2fil(fout, obsglob, fscrunch, subdm, source, ram):
    """Convert a list of CHIME/FRB msgpack files to a filterbank file.

    Parameters
    ----------
    fout : str
        Filterbank file name to write out, '.fil' is appended if missing.
    obsglob : str
        Unix pathname wildcard to .msgpack files to be converted.
    fscrunch : int
        Frequency scrunch factor, or the numbers of channels that will be
        scrunched together into one subband. Needs to be a power of 2.
    subdm : float
        Disperions measure, in pc cm-3, that channels withing a subband will
        be dedispersed to to avoid subband smearing. NB the filterbank file
        wil still be at DM=0.
    source : str
        Name of the source in the observation that will be put in the
        filterbank header.
    ram : float
        Random access memory available, in bytes, on your machine; used to
        calculate the chunk size of file reads. NB '8e9' for 8GB, etc.

    """
    # add extension to filename if it is not already there
    if fout[-4:] != ".fil":
        fout += ".fil"

    msg = glob.glob(obsglob)
    msg.sort(key=chime_intensity.natural_keys)

    if len(msg) == 0:
        print(f"No files found for wildcard '{obsglob}'..")
        return

    # indices of chunk number is different for `.msg` and `.msgpack` files
    _, extension = os.path.splitext(msg[0])
    if extension == ".msg":
        chidx1 = -12
        chidx2 = -4
    if extension == ".msgpack":
        chidx1 = -19
        chidx2 = -11

    # use 50% of availabe RAM; need 3x the msgpack size
    chunk_size = int(0.5 * ram / MSGPACK_SIZE / 3.0)

    limits = range(0, len(msg), chunk_size)
    limits.append(len(msg))

    reads = len(limits) - 1

    for i in range(reads):
        print("Reading {} of {} msgpack chunks..".format(i + 1, reads))

        msg_chunk = msg[limits[i] : limits[i + 1]]

        intensity, weights, dt, df = convert_chunk(
            msg_chunk, fscrunch=fscrunch, subdm=subdm
        )

        if i > 0:
            print(f"Appending to filterbank file '{fout}'..")
            filterbank.append_spectra(outfile, intensity * weights)
        else:
            print(f"Creating filterbank file '{fout}'..")
            outfile = open(fout, "wb")
            fil_header = filterbank_header(
                obsglob,
                dt,
                chime_intensity.ftop,
                chime_intensity.fbottom,
                intensity.shape[0],
                source,
            )
            filterbank.create_filterbank_file(outfile, fil_header, intensity * weights)

    outfile.close()
    return


@click.command()
@click.option(
    "-o",
    "--outfile",
    "fout",
    default="chimefrb.fil",
    type=click.STRING,
    help="Filterbank file name.",
)
@click.option(
    "--obsglob",
    default="./*.msgpack",
    type=click.STRING,
    help="Glob for the observation msgpack files in quotes.",
)
@click.option(
    "--fscrunch",
    default=4,
    type=click.INT,
    help="Frequency scrunch factor, needs to be a power of 2.",
)
@click.option(
    "--subdm",
    default=None,
    type=click.FLOAT,
    help="Before subbanding dedisperse subband channels to this DM.",
)
@click.option(
    "--source",
    default="CHIME/FRB candidate",
    type=click.STRING,
    help="Name of the source that is in the observation.",
)
@click.option(
    "--ram",
    default=8e9,
    type=click.FLOAT,
    help="RAM available, in bytes. 8GB by default.",
)
def runner(fout, obsglob, fscrunch, subdm, source, ram):
    msgpack2fil(fout, obsglob, fscrunch, subdm, source, ram)


if __name__ == "__main__":
    runner()



import logging
import urllib.request

from cfod import DATA_DIR

logging.basicConfig(format="%(levelname)s:%(message)s")
log = logging.getLogger(__name__)


def _download(url: str, destination: str) -> None:
    """
    Download a file from a URL .

    Parameters
    ----------
    url : str
        Url to the data product.
    destination : str
        Save path for the downloaded file.
    """

    try:
        log.debug("Downloading...")
        log.debug(f"URL  : {url}")
        log.debug(f"File : {destination}")
        urllib.request.urlretrieve(url, destination)
    except Exception as error:
        log.warning("Unable to download CHIME/FRB Catalog.")
        raise error


def csv_catalog():
    """
    Download CSV catalog.
    """
    url = "https://storage.googleapis.com/chimefrb-dev.appspot.com/catalog1/chimefrbcat1.csv"
    datapath = DATA_DIR / "catalog.csv"
    destination = datapath.absolute().as_posix()
    _download(url, destination)


def fits_catalog():
    """
    Download the fits catalog .
    """
    url = "https://storage.googleapis.com/chimefrb-dev.appspot.com/catalog1/chimefrbcat1.fits"
    datapath = DATA_DIR / "catalog.fits"
    destination = datapath.absolute().as_posix()
    _download(url, destination)


def data_product(url: str) -> None:
    """
    Download data from a URL.

    Parameters
    ----------
    url : str
        URL for the data file.
    """
    datapath = DATA_DIR / url.split("/")[-1]
    destination = datapath.absolute().as_posix()
    _download(url, destination)



import logging
import os
from pathlib import Path

from cfod.routines import catalogs

logging.basicConfig(format="%(levelname)s:%(message)s")
log = logging.getLogger(__name__)

BASE_DIR: Path = Path(os.path.dirname(os.path.realpath(__file__)))
DATA_DIR: Path = BASE_DIR / "data"
CSV_CATALOG: Path = DATA_DIR / "catalog.csv"
FITS_CATALOG: Path = DATA_DIR / "catalog.fits"

if os.environ.get("DEBUG", False):
    log.setLevel(logging.DEBUG)
    log.debug("Logging Level: Debug")

if not len(list(DATA_DIR.glob("catalog.*"))):
    from cfod.utilities import fetch

    log.debug("Fetching CHIME/FRB Catalogs")
    fetch.csv_catalog()
    fetch.fits_catalog()

if CSV_CATALOG.exists():
    catalog = catalogs.Catalogs(filename=CSV_CATALOG.absolute().as_posix())
elif FITS_CATALOG.exists():
    catalog = catalogs.Catalogs(filename=FITS_CATALOG.absolute().as_posix())
else:
    log.error("Unable to locate CHIME/FRB Catalog.")



"""Functions to handle CHIME/FRB intensity data."""


import numpy as np

from cfod.analysis.intensity import assemble_chunk

# CHIME/FRB Constants

# ADC sampling frequency, in MSPS or Hz
adc_sampling_freq = float(800e6)
# number of samples in the initial FFT
fpga_num_samp_fft = 2048
# parameters for alias sampling in second Nyquist zone
fpga_num_freq = fpga_num_samp_fft / 2
# bin centre of the highest frequency channel
fpga_freq0_mhz = adc_sampling_freq / 1e6
# channel bandwidth
fpga_delta_freq_mhz = -adc_sampling_freq / 2 / fpga_num_freq / 1e6
# top of the highest-frequency channel
# (NB the FGPA-channel around 800 MHz is contaminated by aliasing)
freq_top_mhz = fpga_freq0_mhz - fpga_delta_freq_mhz / 2.0
# bottom of the lowest-frequency channel
freq_bottom_mhz = freq_top_mhz - adc_sampling_freq / 2.0 / 1e6
# bin centres of FPGA channels, in MHz (ordered 800 to 400 MHz)
fpga_freq = np.linspace(
    fpga_freq0_mhz, fpga_freq0_mhz / 2.0, num=round(fpga_num_freq), endpoint=False
)

# X-engine
# upchannelization
l0_upchan_factor = 16
l0_num_frames_sample = 8 * 3
# CHIME/FRB output data parameters
num_channels = fpga_num_freq * l0_upchan_factor
channel_bandwidth_mhz = adc_sampling_freq / 2 / num_channels / 1e6
fpga_frequency_hz = adc_sampling_freq / fpga_num_samp_fft
sampling_time_ms = (
    1.0 / fpga_frequency_hz * l0_upchan_factor * l0_num_frames_sample
) * 1e3
fpga_counts_per_sample = int(sampling_time_ms / 1e3 / (1.0 / fpga_frequency_hz))

# bin centres of L0 channels, in MHz (ordered 400 to 800 MHz)
freq = np.arange(
    freq_bottom_mhz + channel_bandwidth_mhz / 2.0, freq_top_mhz, channel_bandwidth_mhz
)
bandwidth = adc_sampling_freq / 1e6 / 2.0  # MHz
fbottom = freq_bottom_mhz  # MHz
ftop = freq_top_mhz  # MHz
df = channel_bandwidth_mhz  # MHz
dt = sampling_time_ms / 1e3  # s
nchan = num_channels
fpga_counts_per_sample = fpga_counts_per_sample


def unpack_data(fn):
    """
    Unpacks and de-compresses Intensity and Weights from the
    L1 call-back data.
    Parameters
    ----------
    fn : string
        Filename of the function to unpack.
        For CHIME/FRB this has the following format:
        astro_5941664_20180406203904337770_beam0147_00245439_02.msgpack
              event_no_YYYYMMDDHHMMSSssss_beamxxxx_something_binning.msgpack
    Returns
    -------
    intensity : 2D array
        A 2D Intensity array.
    weights : 2D array
        Corresponding weights to the intensity array.
    fpga0 : int
        Start fpga count of that chunk.
    fpgaN : int
        number of fpga counts in that chunk.
    binning : int
        Downsampling of the data from the ringbuffer
    """

    chunk = assemble_chunk.read_msgpack_file(fn)
    intensity, weights = chunk.decode()
    frame0_nano = None
    nrfifreq = None
    rfi_mask = np.ones_like(intensity)
    version = chunk.version
    if version == 2:
        frame0_nano = chunk.frame0_nano
        nrfifreq = chunk.nrfifreq
        rfi_mask = chunk.rfi_mask
    return (
        intensity,
        weights,
        chunk.fpga0,
        chunk.fpgaN,
        chunk.binning,
        frame0_nano,
        nrfifreq,
        rfi_mask,
    )


def unpack_datafiles(fns, downsample=True):
    """
    Unpacks a list of functions from a beam and appends to the
    list of intensities and weights. If mixed binning chunks are provided,
    downsampled chunks are upsampled to the finest time resolution
    (lowest binning) available.
    Parameters
    ----------
    fns : list
        A list of function names from a beam corresponsing
        to different time chunks.
    Return
    ------
    intensities : list
        A list of 2D intensity arrays.
    weights : list
        A list of 2D weight arrays.
    fpga0s : int
        A list start fpga count of those chunks.
    fpgaNs : int
        A list of number of fpga counts in those chunks.
    binning : int
        The downsampling used for the chunks.
        If a mixture of binnings is used, the intensity data is duplicated
        to the finest time resolution.
    """
    intensities = []
    weights = []
    fpga0s = []
    fpgaNs = []
    rfi_masks = []
    bin_list = []
    frame0_nanos = []
    for fn in fns:
        print(fn)
        (
            intensity,
            weight,
            fpga0,
            fpgaN,
            binning,
            frame0_nano,
            nrfifreq,
            rfi_mask,
        ) = unpack_data(fn)
        intensities.append(intensity)
        weights.append(weight)
        fpga0s.append(fpga0)
        fpgaNs.append(fpgaN)
        bin_list.append(binning)
        frame0_nanos.append(frame0_nano)
        rfi_masks.append(rfi_mask)
    print("files read...")
    output_bin = 1

    if (
        len(set(bin_list)) == 1 and bin_list[0] == 1
    ):  # if bin_list has only one unique item and it is 1
        return (
            np.hstack(intensities)[::-1],
            np.hstack(weights)[::-1],
            fpga0s,
            fpgaNs,
            output_bin,
            np.hstack(rfi_masks)[::-1],
            frame0_nanos,
        )
    else:
        print("else condition triggered")
        # get all of the intensities to the same
        # get the output shape:

        # Downsample
        if downsample:
            raise NotImplementedError("data downsampling is not supported.")

        # Upsample
        else:
            total_samples = np.sum(
                [
                    (binning / output_bin) * intensity_chunk.shape[1]
                    for binning, intensity_chunk in zip(bin_list, intensities)
                ]
            )

            output_intensities = np.empty([intensities[0].shape[0], total_samples])
            output_weights = np.empty_like(output_intensities)
            output_rfi_masks = np.empty_like(output_intensities)
            current_idx = 0

            for i in range(len(bin_list)):
                binning = bin_list[i]
                intensity = intensities[i].repeat(binning / output_bin, axis=1)
                chunk_samples = intensity.shape[1]
                output_intensities[
                    :, current_idx : current_idx + chunk_samples
                ] = intensity
                weight = weights[i].repeat(binning / output_bin, axis=1)
                output_weights[:, current_idx : current_idx + chunk_samples] = weight
                current_idx += chunk_samples
                rfi_mask = rfi_masks[i].repeat(binning / output_bin, axis=1)
                output_rfi_masks[
                    :, current_idx : current_idx + chunk_samples
                ] = rfi_mask
                current_idx += chunk_samples
        print("all is well")
        return (
            output_intensities,
            output_weights,
            fpga0s,
            fpgaNs,
            output_bin,
            output_rfi_masks,
            frame0_nanos,
        )


# sort files in natural order
def atoi(text: str):
    return int(text) if text.isdigit() else text


def natural_keys(text):
    import re

    """alist.sort(key=natural_keys) sorts in human order"""
    return [atoi(c) for c in re.split(r"(\d+)", text)]



import logging
from typing import Optional

import h5py as h5

from cfod.analysis import localization

logging.basicConfig(format="%(levelname)s:%(message)s")
log = logging.getLogger(__name__)


class Localizer:
    def __init__(self, filename: str):
        """
        Initialize localization object.

        Parameters
        ----------
        filename : str
            Path to localization h5 localization data.
        """
        self.filename = filename
        self.datafile = h5.File(filename, "r")

    def describe(self, group: str, attribute: Optional[str], recurse: bool = False):
        """
        Describes a localization data.

        Parameters
        ----------
        group : str
            Name of the root group
        attribute : Optional[str]
            Name of the attribute
        recurse : bool, optional
            by default False

        Example
        -------
        >>> describle("/healpix", "comments")
        """
        if attribute:
            localization.describe(
                group=self.datafile[group].attrs[attribute], recurse=recurse
            )
        else:
            localization.describe(group=self.datafile[group], recurse=recurse)

    def plot(self):
        """
        Plot the localization data.
        """
        localization.plot(data=self.datafile)

    def coutour_plot(self):
        """
        Draw the localization coutour plot
        """
        localization.countours(data=self.datafile)



import h5py as h5
import matplotlib.pyplot as plt
import numpy as np


def describe(group, recurse=False):
    """
    Describes a HDF5 group object

    Parameters
    ----------
    group : [type]
        [description]
    recurse : bool, optional
        [description], by default False
    """

    print(group.name)

    # First print header-like attributes (if exist)
    if group.attrs:
        print("\n  attrs: {")
    for key, value in group.attrs.items():
        if key in ["comments", "history"]:
            print("    %s:" % key)
            for line in value:
                print("      " + str(line))
        else:
            print("    %s:" % key, value)
    if group.attrs:
        print("  }")

    # Then print constituent groups & datasets
    print()
    for key, value in group.items():
        if isinstance(value, h5.Group):
            if recurse:
                print("-" * 60)
                describe(value, True)
            else:
                print("  " + key + "/")
        else:
            print("  " + key + ":", value.shape, value.dtype)
    print()


def plot(data):
    hdr = data["projection"].attrs
    CL = data["projection/data"][:]
    extent = (
        np.array(
            [-hdr["xsize"] / 2, hdr["xsize"] / 2, -hdr["ysize"] / 2, hdr["ysize"] / 2]
        )
        * hdr["reso"]
        / 60
    )

    plt.rc("font", family="serif", size=14)
    plt.figure(figsize=(10, 4))

    # Note: RA increases to the left!
    im = plt.imshow(CL, vmin=0, origin="lower", extent=extent, cmap="magma")
    plt.contour(
        CL,
        levels=[0.68, 0.95],
        linestyles=["-", "--"],
        colors="k",
        linewidths=2,
        extent=extent,
    )

    plt.colorbar(
        im, pad=0.25, shrink=0.4, orientation="horizontal", label="Confidence Level"
    )
    plt.arrow(2.4, -0.4, 0, 0.2, head_width=0.04, color="k")
    plt.text(2.39, -0.1, "N", ha="center", size=10)
    plt.arrow(2.4, -0.4, -0.2, 0.0, head_width=0.04, color="k")
    plt.text(2.1, -0.4, "E", va="center", ha="right", size=10)
    plt.title("Centered @ {:.3f}, {:.2f}".format(hdr["clon"], hdr["clat"]))
    plt.xlabel("dx (deg)")
    plt.ylabel("dy (deg)")


def countours(data):
    # example 0: getting points
    ra, dec = data["contours/68/A"]

    # example 2: plotting contours
    plt.figure(figsize=(10, 2))

    for name, contour in data["contours/68"].items():
        contour = contour[:]
        plt.plot(*contour)
        plt.plot(*contour.mean(1), "wo", mec="k", ms=20, alpha=0.5)
        plt.text(*contour.mean(1), s=name, ha="center", va="center")
    for contour in data["contours/95"].values():
        plt.plot(*contour[:], "--")

    plt.xlim(*plt.xlim()[::-1])
    plt.xlabel("R.A. (deg)")
    plt.ylabel("Dec. (deg)")



import logging
from typing import Optional

import numpy as np

from cfod.analysis import waterfall

logging.basicConfig(format="%(levelname)s:%(message)s")
log = logging.getLogger(__name__)


class Waterfaller:
    def __init__(self, filename: str) -> None:
        """
        Initialize the Waterfaller.

        Parameters
        ----------
        filename : str
            h5 file, container the CHIME/FRB waterfall data.
        """
        self.filename = filename
        self.datafile = waterfall.data(filename=self.filename)
        self._unpack()

    def _unpack(self):
        """
        Unpack the attributes of the datafile.
        """
        self.data = self.datafile["frb"]
        self.eventname = self.datafile.attrs["tns_name"].decode()
        self.wfall = self.datafile["wfall"][:]
        self.model_wfall = self.datafile["model_wfall"][:]
        self.plot_time = self.datafile["plot_time"][:]
        self.plot_freq = self.datafile["plot_freq"][:]
        self.ts = self.datafile["ts"][:]
        self.model_ts = self.datafile["model_ts"][:]
        self.spec = self.datafile["spec"][:]
        self.model_spec = self.datafile["model_spec"][:]
        self.extent = self.datafile["extent"][:]
        self.dm = self.datafile.attrs["dm"][()]
        self.scatterfit = self.datafile.attrs["scatterfit"][()]
        self.cal_obs_date = self.datafile.attrs["calibration_observation_date"].decode()
        self.cal_source_name = self.datafile.attrs["calibration_source_name"].decode()
        self.cal_wfall = self.datafile["calibrated_wfall"][:]
        self.dt = np.median(np.diff(self.plot_time))

    @property
    def contents(self) -> list:
        """
        Return the contents of the datafile as a list of strings .

        Returns
        -------
        list
        """
        return waterfall.contents(self.datafile)

    @property
    def metadata(self) -> list:
        """
        Return the metadata for the datafile .

        Returns
        -------
        list
        """
        return waterfall.metadata(self.datafile)

    @property
    def data_arrays(self) -> list:
        """
        List of data arrays in the data file .

        Returns
        -------
        list
        """
        return waterfall.data_arrays(self.datafile)

    def plot(self, savepath: Optional[str] = None):
        """
        Plot the spectrum.

        Parameters
        ----------
        savepath : Optional[str], optional
            If provided, plot saved as {savepath}/{filename}.png, by default None
        """
        log.info("Removing Remaining RFI")
        ts, model_ts = waterfall.remove_rfi(self.spec, self.wfall, self.model_wfall)
        log.info("Determining the peaks and SNR of the pulse")
        peak, width, snr = waterfall.find_burst(ts)
        log.info(f"Peak: {peak}, Width: {width*self.dt} ms, SNR: {snr}")
        wfall = waterfall.bin_freq_channels(self.wfall, fbin_factor=16)
        log.info("Plotting Dynamic Spectrum")
        waterfall.plot(
            ts=ts,
            extent=self.extent,
            plot_time=self.plot_time,
            dt=self.dt,
            wfall=wfall,
            spec=self.spec,
            plot_freq=self.plot_freq,
            scatterfit=self.scatterfit,
            model_spec=self.model_spec,
            model_ts=model_ts,
            peak=peak,
            width=width,
            snr=snr,
            eventname=self.eventname,
            dm=self.dm,
            savepath=savepath,
        )

    def cal_plot(self, savepath: Optional[str] = None):
        """
        Plot the calibrated spectrum.

        Parameters
        ----------
        savepath : Optional[str], optional
            If provided, plot saved as {savepath}/{filename}.png, by default None
        """
        waterfall.cal_plot(
            cal_wfall=self.cal_wfall,
            dt=self.dt,
            extent=self.extent,
            eventname=self.eventname,
            cal_source_name=self.cal_source_name,
            cal_obs_date=self.cal_obs_date,
            savepath=savepath,
        )



""" Creates a HEALPix map from CHIME/FRB Exposure Map."""
import logging

import astropy.units as u
import healpy as hp
import numpy as np
from astropy.coordinates import SkyCoord

log = logging.getLogger(__name__)


def render(filename: str, save: bool = False, debug: bool = False) -> None:
    """
    Renders CHIME/FRB Exposure Map in Mollweide projection.

    Parameters
    ----------
    filename : str
        Filename of the exposure map.
    debug : bool, optional
        Enables debug logging, by default False
    """
    if debug:
        log.setLevel(logging.DEBUG)
    with np.load(filename) as data:
        exposure = data["exposure"]
        log.debug(f"{filename} opened.")

    # Setting parameters for map resolution
    # Spatial
    nside = 4096
    npix = hp.nside2npix(nside)

    # Temporal
    t_res = 4

    # Initializing a healpy map
    hpxmap = np.zeros(npix, dtype=np.float64)
    hpxmap[0 : len(exposure)] += t_res * exposure / (3600.0)  # seconds to hours
    hpxmap[hpxmap == 0] = hp.UNSEEN  # masking pixels with zero exposure

    # Plotting
    hp.mollview(hpxmap, coord=["C", "G"], norm="log", unit="Hours")
    # Check exposure time in hours for R1 repeater
    coord = SkyCoord("05:31:58.70", "+33:08:52.5", frame="icrs", unit=u.deg)
    log.info(
        "Exposure (in hours): %.2f"
        % hpxmap[hp.ang2pix(nside, coord.ra.deg, coord.dec.deg, lonlat=True)]
    )

    # Obtaining a lower resolution map ###
    nside_out = 1024
    log.info(
        "Resolution of new map : %.2f arcmin" % (hp.nside2resol(nside_out, arcmin=True))
    )
    # Degrade healpix resolution to nside_out
    hpxmap_dg = hp.ud_grade(hpxmap, nside_out)
    hp.mollview(hpxmap_dg, coord=["C", "G"], norm="log", unit="Hours")



"""A module to write filterbank files. Slightly modified version of:
https://github.com/scottransom/presto/blob/master/lib/python/filterbank.py
by Patrick Lazarus.

Ziggy Pleunis, ziggy.pleunis@physics.mcgill.ca

"""

import numpy as np

from cfod.analysis.filterbank import sigproc


def create_filterbank_file(outfile, header, spectra=None, nbits=32, verbose=False):
    """Write filterbank header and spectra to file.

    Parameters
    ----------
    outfile : file
        The filterbank file.
    header : dict
        Dictionary of header parameters and values.
    spectra : Spectra
        Spectra to write to file.
        (Default: don't write any spectra, i.e. write out header only.)
    nbits : int
        The number of bits per sample of the filterbank file. This value
        always overrides the value in the header dictionary.
        (Default: 32, i.e. each sample is a 32-bit float.)
    verbose : bool
        If `True`, be verbose.
        (Default: be quiet.)

    """
    dtype = get_dtype(nbits)
    header["nbits"] = nbits

    if is_float(nbits):
        tinfo = np.finfo(dtype)
    else:
        tinfo = np.iinfo(dtype)
    dtype_min = tinfo.min
    dtype_max = tinfo.max

    outfile.write(sigproc.addto_hdr("HEADER_START", None))

    for parameter in list(header.keys()):
        if parameter not in sigproc.header_params:
            # only add recognized parameters
            continue
        if verbose:
            print(f"Writing header parameter '{parameter}'")
        value = header[parameter]
        outfile.write(sigproc.addto_hdr(parameter, value))

    outfile.write(sigproc.addto_hdr("HEADER_END", None))

    if spectra is not None:
        # CHIME/FRB spectra have shape (channels, time_samples) so tranpose
        data = spectra.T
        nspec, nchans = data.shape
        # first channel in filterbank should be highest frequency, so flip band
        fil_data = np.fliplr(data).ravel()
        np.clip(fil_data, dtype_min, dtype_max, out=fil_data)
        fil_data.tofile(outfile, format=dtype)


def append_spectra(outfile, spectra, nbits=32, verbose=False):
    """Append filterbank spectra to file.

    Parameters
    ----------
    outfile : file
        The filterbank file.
    spectra : Spectra
        Spectra to append to file.
    nbits : int
        The number of bits per sample of the filterbank file. This value
        always overrides the value in the header dictionary.
        (Default: 32, i.e. each sample is a 32-bit float.)
    verbose : bool
        If `True`, be verbose.
        (Default: be quiet.)

    """
    dtype = get_dtype(nbits)
    if is_float(nbits):
        tinfo = np.finfo(dtype)
    else:
        tinfo = np.iinfo(dtype)
    dtype_min = tinfo.min
    dtype_max = tinfo.max

    # CHIME/FRB spectra have shape (channels, time_samples) so tranpose
    data = spectra.T
    nspec, nchans = data.shape
    # first channel in filterbank should be highest frequency so flip band
    fil_data = np.fliplr(data).ravel()
    np.clip(fil_data, dtype_min, dtype_max, out=fil_data)
    # outfile.seek(0, os.SEEK_END)
    outfile.write(fil_data.astype(dtype))


def is_float(nbits):
    """For a given number of bits per sample return `True` if it
    corresponds to floating-point samples in filterbank files.

    Parameters
    ----------
    nbits: int
        Number of bits per sample, as recorded in the filterbank file's
        header.

    Returns
    -------
    isfloat : bool
        `True`, if `nbits` indicates the data in the file are encoded
        as floats.

    """
    check_nbits(nbits)
    if nbits == 32:
        return True
    else:
        return False


def check_nbits(nbits):
    """Given a number of bits per sample check to make sure
    `filterbank.py` can cope with it.

    Parameters
    ----------
    nbits : int
        Number of bits per sample, as recorded in the filterbank file's
        header.

    Raises
    ------
    ValueError
        If `filterbank.py` cannot cope.

    """
    if nbits not in [32, 16, 8]:
        raise ValueError(
            "`filterbank.py` only supports files with 8- or "
            "16-bit integers, or 32-bit floats "
            "(nbits provided: {})!".format(nbits)
        )


def get_dtype(nbits):
    """For a given number of bits per sample return a numpy-recognized
    dtype.

    Parameters
    ----------
    nbits : int
        Number of bits per sample, as recorded in the filterbank file's
        header.

    Returns
    -------
    dtype : dtype
        A numpy dtype string.

    """
    check_nbits(nbits)

    if is_float(nbits):
        dtype = f"float{nbits}"
    else:
        dtype = f"uint%d"

    return dtype



"""CHIME/FRB Catalog."""
import logging
from pathlib import Path
from typing import Optional

from astropy.io import fits

from cfod.utilities import parse

logging.basicConfig(format="%(levelname)s:%(message)s")
log = logging.getLogger(__name__)


class Catalogs:
    """CHIME/FRB Catalogs."""

    def __init__(self, filename: str, debug: bool = False):
        """
        CHIME/FRB Catalog.

        Parameters
        ----------
        filename : str
            Name of the catalog file.
        """
        if debug:
            log.setLevel(logging.DEBUG)
        self.filename: str = filename
        self.format: Optional[str] = None
        self._filecheck()

    def _filecheck(self, filetypes: list = [".csv", ".fits"]):
        """
        Check if the file exists and is of valid format.

        Parameters
        ----------
        filetypes : list, optional
            Acceptable tile types, by default [".csv", ".fits"]

        Raises
        ------
        FileNotFoundError
            Raised when filename does not exist.
        ValueError
            Raised when filename is not of valid format.
        """
        exists = Path(self.filename).exists()
        suffix = Path(self.filename).suffix
        log.debug(f"Loading Catalog: {self.filename}")
        if not exists:
            raise FileNotFoundError(f"{self.filename} does not exist.")
        if suffix not in filetypes:
            raise ValueError(f"{self.filename} does not have type {filetypes}.")
        self.format = suffix

    def as_dict(self) -> dict:
        """
        Return the CSV data as a dict.

        Note
        ----
        The format of the dictionary is,
            {
                FRB_NAME : FRB_PROPERTIES
            }

        Returns
        -------
        dict
        """
        self._filecheck(filetypes=[".csv"])
        return parse.csv_to_dict(self.filename)

    def as_json(self) -> str:
        """
        Return the CSV as a JSON string, follows the dictionary format.

        Returns
        -------
        str
        """
        self._filecheck(filetypes=[".csv"])
        return parse.csv_to_json(self.filename)

    def as_list(self) -> list:
        """
        Return a list of the CSV data .

        Returns
        -------
        list
        """
        self._filecheck(filetypes=[".csv"])
        return parse.csv_to_list(filename=self.filename)

    def as_dataframe(self):
        """
        Return the fits data as a pandas dataframe.

        Returns
        -------
        pandas.Dataframe
        """
        self._filecheck(filetypes=[".csv"])
        return parse.fits_to_dataframe(filename=self.filename)

    def as_fits(self) -> fits.HDUList:
        """
        Return a fits HDUList object for the catalog.

        Returns
        -------
        fits.HDUList
            CHIME/FRB Catalog
        """
        self._filecheck(filetypes=[".fits"])
        return fits.open(self.filename)

    def as_ndarray(self) -> NotImplementedError:
        """
        Would return a np.ndarray.

        Raises
        ------
        NotImplementedError
        """
        raise NotImplementedError("Currently not implemented. PRs are welcomed!")



"""Sigproc header definitions and tools. Slightly modified version of:
https://github.com/scottransom/presto/blob/master/lib/python/sigproc.py
by Patrick Lazarus.

Ziggy Pleunis, ziggy.pleunis@physics.mcgill.ca

"""

import struct

header_params = {
    "HEADER_START": "flag",
    "telescope_id": "i",
    "machine_id": "i",
    "data_type": "i",
    "rawdatafile": "str",
    "source_name": "str",
    "barycentric": "i",
    "pulsarcentric": "i",
    "az_start": "d",
    "za_start": "d",
    "src_raj": "d",
    "src_dej": "d",
    "tstart": "d",
    "tsamp": "d",
    "nbits": "i",
    "nsamples": "i",
    "nbeams": "i",
    "ibeam": "i",
    "fch1": "d",
    "foff": "d",
    "FREQUENCY_START": "flag",
    "fchannel": "d",
    "FREQUENCY_END": "flag",
    "nchans": "i",
    "nifs": "i",
    "refdm": "d",
    "period": "d",
    "npuls": "q",
    "nbins": "i",
    "HEADER_END": "flag",
}


def prep_string(string):
    return struct.pack("i", len(string)) + string


def prep_double(name, value):
    return prep_string(name) + struct.pack("d", float(value))


def prep_int(name, value):
    return prep_string(name) + struct.pack("i", int(value))


def addto_hdr(parameter, value):
    """Prepare parameter and value for writing to binary file."""
    if header_params[parameter] == "d":
        return prep_double(parameter, value)
    elif header_params[parameter] == "i":
        return prep_int(parameter, value)
    elif header_params[parameter] == "str":
        return prep_string(parameter) + prep_string(value)
    elif header_params[parameter] == "flag":
        return prep_string(parameter)
    else:
        print(f"WARNING key '{parameter}' is unknown!")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cfod',
 'cfod.analysis',
 'cfod.analysis.filterbank',
 'cfod.analysis.intensity',
 'cfod.data',
 'cfod.routines',
 'cfod.utilities']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.0.1,<9.0.0',
 'h5py>=3.2.1,<4.0.0',
 'healpy>=1.14.0,<2.0.0',
 'matplotlib>=3.4.2,<4.0.0',
 'msgpack-python>=0.5.6,<0.6.0',
 'numpy>=1.20.3,<2.0.0',
 'scipy>=1.6.3,<2.0.0']

extras_require = \
{'pandas': ['pandas>=1.2.4,<2.0.0']}

entry_points = \
{'console_scripts': ['msgpack2fil = cfod.routines.msgpack2fil:runner']}

setup_kwargs = {
    'name': 'cfod',
    'version': '2021.6.4',
    'description': 'CHIME FRB Open Data',
    'long_description': '<h1 align="center">\n  <br>\n  <a href="https://chime-frb-open-data.github.io"><img src="https://github.com/chime-frb-open-data/chime-frb-open-data.github.io/blob/79d7c2d574a6c849125583395f5442333630222d/docs/static/chime-frb-logo.png" alt="" width="25%"></a>\n  <br>\n  Utililes for CHIME/FRB Open Data Releases.\n  <br>\n</h1>\n\n\n## Installation\n```\npip install --user cfod\n\n# To install with pandas support,\npip install --user cfod[pandas]\n```\n\n## Documentation\nCheck out the user documentation, [here](https://chime-frb-open-data.github.io/)\n\n\n## Developer\n```\n# cfod uses poetry for package management and virtualenv management\npip install poetry\n\ngit clone git@github.com:chime-frb-open-data/chime-frb-open-data.git\ncd chime-frb-open-data\n\n# Install git-commit hook\npoetry run pre-commit install\n\n# Make changes to the code and open a PR\n```\n\n## Removal\n```\npip uninstall cfod\n```\n\n\n<p align="center">\n  <a href="Some Love">\n    <img src="https://forthebadge.com/images/badges/built-with-love.svg">\n  </a>\n</p>\n',
    'author': 'Shiny Brar',
    'author_email': 'charanjotbrar@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://chime-frb-open-data.github.io/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7.1,<3.10',
}


setup(**setup_kwargs)



from typing import Optional, Tuple

import h5py
import matplotlib.gridspec as gridspec
import matplotlib.pyplot as plt
import numpy as np
import scipy.signal


def data(filename: str) -> h5py.File:
    """
    Create a h5py File object for the given filename .

    Parameters
    ----------
    filename : str
        Path to the filename.

    Returns
    -------
    h5py.File
        h5py File object.
    """
    return h5py.File(filename, "r")


def contents(data: h5py.File) -> list:
    """
    Returns the list of the contents of a h5py file.

    Parameters
    ----------
    data : h5py.File

    Returns
    -------
    list
    """
    return list(data.keys())


def metadata(data: h5py.File) -> list:
    """
    Returns the metadata for the given CHIME/FRB Waterfall.

    Parameters
    ----------
    data : h5py.File

    Returns
    -------
    list
    """
    return list(data["frb"].attrs)


def data_arrays(data: h5py.File) -> list:
    """
    List of all the FRB data arrays in a file .

    Parameters
    ----------
    data : h5py.File

    Returns
    -------
    list
    """
    return list(data["frb"].keys())


def remove_rfi(
    spec: np.ndarray, wfall: np.ndarray, model_wfall: np.ndarray
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Set any frequency channel that has a higher variance than
    the mean variance (averaged across all frequency channels)
    to a np.nan.

    Parameters
    ----------
    spec : np.ndarray
    wfall : np.ndarray
    model_wfall : np.ndarray

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
    """
    q1 = np.nanquantile(spec, 0.25)
    q3 = np.nanquantile(spec, 0.75)
    iqr = q3 - q1

    # additional masking of channels with RFI
    rfi_masking_var_factor = 3

    channel_variance = np.nanvar(wfall, axis=1)
    mean_channel_variance = np.nanmean(channel_variance)

    with np.errstate(invalid="ignore"):
        rfi_mask = (
            (channel_variance > rfi_masking_var_factor * mean_channel_variance)
            | (spec[::-1] < q1 - 1.5 * iqr)
            | (spec[::-1] > q3 + 1.5 * iqr)
        )
    wfall[rfi_mask, ...] = np.nan
    model_wfall[rfi_mask, ...] = np.nan
    spec[rfi_mask[::-1]] = np.nan

    # remake time-series after RFI masking
    ts = np.nansum(wfall, axis=0)
    model_ts = np.nansum(model_wfall, axis=0)

    return ts, model_ts


def boxcar_kernel(width: int) -> np.ndarray:
    """
    Returns the boxcar kernel of given width normalized by
    sqrt(width) for S/N reasons.

    Parameters
    ----------
    width : int
        Width of the boxcar.
    Returns
    -------
    boxcar : np.ndarray
        Boxcar of width `width` normalized by sqrt(width).
    """
    width = int(round(width, 0))
    return np.ones(width, dtype="float32") / np.sqrt(width)


def find_burst(ts, min_width=1, max_width=128) -> Tuple[int, int, float]:
    """
    Find burst peak and width using boxcar convolution.

    Parameters
    ----------
    ts : array_like
        Time-series.
    min_width : int, optional
        Minimum width to search from, in number of time samples.
        1 by default.
    max_width : int, optional
        Maximum width to search up to, in number of time samples.
        128 by default.
    plot : bool, optional
        If True, show figure to summarize burst finding results.
    Returns
    -------
    peak : int
        Index of the peak of the burst in the time-series.
    width : int
        Width of the burst in number of samples.
    snr : float
        S/N of the burst.

    """
    min_width = int(min_width)
    max_width = int(max_width)

    # do not search widths bigger than timeseries
    widths = list(range(min_width, min(max_width + 1, len(ts) - 2)))

    # envelope finding
    snrs = np.empty_like(widths, dtype=float)
    peaks = np.empty_like(widths, dtype=int)

    for i in range(len(widths)):
        convolved = scipy.signal.convolve(ts, boxcar_kernel(widths[i]), mode="same")
        peaks[i] = np.nanargmax(convolved)
        snrs[i] = convolved[peaks[i]]

    best_idx = np.nanargmax(snrs)

    return peaks[best_idx], widths[best_idx], snrs[best_idx]


def bin_freq_channels(data: np.ndarray, fbin_factor: int = 4) -> np.ndarray:
    """
    Bin Data. Sum groups of frequency channels

    Parameters
    ----------
    data : [2D ndarray]
    num_tbin : int, optional
        [number of time samples to bins], by default 100
    fbin_factor : int, optional
        [factor by which frequencies to be binned], by default 1

    Returns
    -------
    [ndarray]

    Raises
    ------
    ValueError
        [frequency binning factor should be even]
    """

    num_chan = data.shape[0]
    if num_chan % fbin_factor != 0:
        raise ValueError("frequency binning factor `fbin_factor` should be even")
    data = np.nanmean(
        data.reshape((num_chan // fbin_factor, fbin_factor) + data.shape[1:]), axis=1
    )
    return data


def plot(
    ts: np.ndarray,
    extent: np.ndarray,
    plot_time: np.ndarray,
    dt: np.float64,
    wfall: np.ndarray,
    spec: np.ndarray,
    plot_freq: np.ndarray,
    scatterfit: bool,
    model_spec: np.ndarray,
    model_ts: np.ndarray,
    peak: int,
    width: int,
    snr: float,
    eventname: str,
    dm: float,
    savepath: Optional[str] = None,
) -> None:
    """
    Plot the time series of the given CHIME/FRB Event Data.


    Parameters
    ----------
    ts : np.ndarray
    extent : np.ndarray
    plot_time : np.ndarray
    dt : np.float64
    wfall : np.ndarray
    spec : np.ndarray
    plot_freq : np.ndarray
    scatterfit : bool
    model_spec : np.ndarray
    model_ts : np.ndarray
    peak : int
    width : int
    snr : float
    eventname : str
    dm : float
    save : bool, optional
    """
    fig = plt.figure(figsize=(6, 6))

    # Set up the image grid
    gs = gridspec.GridSpec(
        ncols=2,
        nrows=2,
        figure=fig,
        width_ratios=[3, 1],
        height_ratios=[1, 3],
        hspace=0.0,
        wspace=0.0,
    )

    data_im = plt.subplot(gs[2])
    data_ts = plt.subplot(gs[0], sharex=data_im)
    data_spec = plt.subplot(gs[3], sharey=data_im)

    # time stamps relative to the peak
    peak_idx = np.argmax(ts)
    extent[0] = extent[0] - plot_time[peak_idx]
    extent[1] = extent[1] - plot_time[peak_idx]
    plot_time -= plot_time[peak_idx]

    # prepare time-series for histogramming
    plot_time -= dt / 2.0
    plot_time = np.append(plot_time, plot_time[-1] + dt)

    cmap = plt.cm.viridis

    # plot dynamic spectrum
    wfall[np.isnan(wfall)] = np.nanmedian(wfall)
    # replace nans in the data with the data median
    # use standard deviation of residuals to set color scale
    vmin = np.nanpercentile(wfall, 1)
    vmax = np.nanpercentile(wfall, 99)

    data_im.imshow(
        wfall,
        aspect="auto",
        interpolation="none",
        extent=extent,
        vmin=vmin,
        vmax=vmax,
        cmap=cmap,
    )

    # plot time-series
    data_ts.plot(
        plot_time, np.append(ts, ts[-1]), color="tab:gray", drawstyle="steps-post"
    )

    # plot spectrum
    data_spec.plot(spec, plot_freq, color="tab:gray")

    # plot model time-series and spectrum
    if scatterfit:
        data_spec.plot(model_spec, plot_freq, color=cmap(0.25))
        data_ts.plot(
            plot_time,
            np.append(model_ts, model_ts[-1]),
            color=cmap(0.25),
            drawstyle="steps-post",
            lw=2,
        )
    else:
        data_spec.plot(model_spec, plot_freq, color=cmap(0.5))
        data_ts.plot(
            plot_time,
            np.append(model_ts, model_ts[-1]),
            color=cmap(0.5),
            drawstyle="steps-post",
            lw=1,
        )

    # BEautify plot
    # remove some labels and ticks for neatness
    plt.setp(data_ts.get_xticklabels(), visible=False)
    data_ts.set_yticklabels([], visible=True)
    data_ts.set_yticks([])
    data_ts.set_xlim(extent[0], extent[1])
    plt.setp(data_spec.get_yticklabels(), visible=False)
    data_spec.set_xticklabels([], visible=True)
    data_spec.set_xticks([])
    data_spec.set_ylim(extent[2], extent[3])
    plt.setp(data_im.get_xticklabels(), fontsize=9)
    plt.setp(data_im.get_yticklabels(), fontsize=9)

    # highlighting the width of the pulse
    data_ts.axvspan(
        max(plot_time.min(), plot_time[peak] + 0.5 * dt - (0.5 * width) * dt),
        min(plot_time.max(), plot_time[peak] + 0.5 * dt + (0.5 * width) * dt),
        facecolor="tab:blue",
        edgecolor=None,
        alpha=0.1,
    )

    # add event ID and DM labels
    xlim = data_ts.get_xlim()
    ylim = data_ts.get_ylim()

    # add 20% extra white space at the top
    span = np.abs(ylim[1]) + np.abs(ylim[0])
    data_ts.set_ylim(ylim[0], ylim[1] + 0.2 * span)
    ylim = data_ts.get_ylim()

    ypos = (ylim[1] - ylim[0]) * 0.9 + ylim[0]
    xpos = (xlim[1] - xlim[0]) * 0.98 + extent[0]
    data_ts.text(
        xpos,
        ypos,
        f"{eventname}\nDM: {dm:.1f} pc/cc\nSNR: {snr:.2f}",
        ha="right",
        va="top",
        fontsize=9,
    )

    data_im.locator_params(axis="x", min_n_ticks=3)
    data_im.set_yticks([400, 500, 600, 700, 800])
    data_im.set_ylabel("Frequency [MHz]", fontsize=9)
    data_im.set_xlabel("Time [ms]", fontsize=9)
    if savepath:
        plt.savefig(f"{savepath}/{eventname}_wfall.png", dpi=300, bbox_inches="tight")


def cal_plot(
    cal_wfall: np.ndarray,
    dt: np.float64,
    extent: np.ndarray,
    eventname: str,
    cal_source_name: str,
    cal_obs_date: str,
    savepath: Optional[str] = None,
):
    """
    Plot the calibrated CHIME/FRB Event Data.

    Parameters
    ----------
    cal_wfall : np.ndarray
    dt : np.float64
    extent : np.ndarray
    eventname : str
    cal_source_name : str
    cal_obs_date : str
    save : bool, optional
    """
    cal_ts = np.nanmean(cal_wfall, axis=0)
    cal_wfall[np.isnan(cal_wfall)] = np.nanmedian(
        cal_wfall
    )  # replace nans in the data with the data median
    # bin frequency channels such that we have 16,384/16 = 1024 frequency channels
    cal_wfall = bin_freq_channels(cal_wfall, 16)
    vmin = np.nanpercentile(cal_wfall, 1)
    vmax = np.nanpercentile(cal_wfall, 99)

    times = np.arange(len(cal_ts)) * dt
    peak_idx = np.argmax(cal_ts)
    times -= times[peak_idx]
    times -= dt / 2.0

    extent[0] = times[0]
    extent[1] = times[-1]

    fig = plt.figure(figsize=(5, 5), constrained_layout=True)
    layout = """
    A
    C
    """
    ax_dict = fig.subplot_mosaic(layout)
    ax_dict["A"].imshow(cal_wfall, aspect="auto", vmin=vmin, vmax=vmax, extent=extent)
    ax_dict["A"].set_title(
        f"Waterfall of {eventname} \n Calibrated to {cal_source_name} on {cal_obs_date}"
    )
    ax_dict["A"].set_yticks([400, 500, 600, 700, 800])
    ax_dict["C"].plot(times, cal_ts, drawstyle="steps-post")
    ax_dict["C"].set_xlabel("Time [ms]")
    ax_dict["C"].set_title(
        f"Time Series of {eventname} \nCalibrated to {cal_source_name} on {cal_obs_date} \
            \n Peak flux = {cal_ts[peak_idx]:.3f} Jy"
    )
    ax_dict["A"].set_ylabel("Frequency [MHz]")
    ax_dict["C"].set_ylabel("Flux [Jy]")
    if savepath:
        plt.savefig(
            f"{savepath}/{eventname}_cal_wfall.png", dpi=300, bbox_inches="tight"
        )
